//-------------------------------------------------------------------------
// play-v6 - music player for ATmega328P @ 16 MHz
// audio PWM control
// by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

#include "pwm.h"
#include <stdint.h>
#include <avr/io.h>

//-------------------------------------------------------------------------
// local functions

static void InitTimer0(void);
static void InitTimer1(void);
static void InitTimer2(void);
static void WaitPWMCycles(uint16_t cycles);

//-------------------------------------------------------------------------
// init audio PWM: init and start PWM timers 0,1,2 (in this order)
// each timer controls 2 PWM outputs via its 2 comparators
// (timer 1 is used as reference)
// insert a delay of PWM_STAGGER_CYCLES between timers
// (cycles are used instead of us, to fit cycle count of macro in Play())

void InitAudioPWM(void)
{
    //initialize timers, keep them stopped (order does not matter here)
    InitTimer0();
    InitTimer1();
    InitTimer2();

    //start timers in 0,1,2 order with given inter-timer delay
    TCCR0B |= _BV(CS00); //start timer 0 - acts as time reference

    //wait (5 cycles will be used to |= set timer 1)
    __builtin_avr_delay_cycles(PWM_STAGGER_CYCLES - 5);

    TCCR1B |= _BV(CS10); //start timer 1

    //wait (5 cycles will be used to set timer 2)
    __builtin_avr_delay_cycles(PWM_STAGGER_CYCLES - 5);

    TCCR2B |= _BV(CS20); //start timer 2
}

//-------------------------------------------------------------------------
// ramp up PWM output to 'silence' level in given time to reduce turn-on noise
// note: no need for precision; time is approximate, especially for small values

void RampUpToSilence(void)
{
    //do nothing in cycle counting test mode
    #if (! TEST_MODE_NOSYNC)

    //number of cycles to wait between ramp-up increments (6 is the number of PWMs)
    const uint16_t delay_cycles = MS_TO_PWM_CYCLES(RAMP_UP_TIME_MS) / ZERO_LEVEL / 6;

    //increment PWM value one-by-one to analog zero
    uint8_t v;
    for (v = 1; v < ZERO_LEVEL; ++v)
    {
        //write new value to all PWMs, one at a time to minimize audible noise
        PWM1Set = v;
        WaitPWMCycles(delay_cycles);
        PWM2Set = v;
        WaitPWMCycles(delay_cycles);
        PWM3Set = v;
        WaitPWMCycles(delay_cycles);
        PWM4Set = v;
        WaitPWMCycles(delay_cycles);
        PWM5Set = v;
        WaitPWMCycles(delay_cycles);
        PWM6Set = v;
        WaitPWMCycles(delay_cycles);
    }

    #endif //(! TEST_MODE_NOSYNC)
}

//-------------------------------------------------------------------------
// setup timer 0 for fast PWM, TOP = 0xff (cycle=256)
// set zero output, enable output on PD6, PD5
// leave timer stopped (no input clock), set TCCR0B |= _BV(CS00) to start

static void InitTimer0(void)
{
    //do not use interrupts
    TIMSK0 = 0x00;
    //no input clock, timer stopped
    TCCR0B = 0x00;
    //reset timer
    TCNT0 = 0x00;

    //fast PWM, count=256 (mode 3), no prescaler,
    //enable compare outputs OC0A, OC0B in non-inverting mode
    TCCR0A = _BV(COM0A1) | _BV(COM0B1) | _BV(WGM01) | _BV(WGM00);

    //set minimum pulse width for both outputs
    OCR0A = 0x00;
    OCR0B = 0x00;

    //enable PWM output on PD6, PD5
    DDRD |= (_BV(DDD6) | _BV(DDD5));

    //(timer is stopped and reset)
}

//-------------------------------------------------------------------------
// setup timer 1 for fast PWM, TOP = 0xff (cycle=256)
// set zero output, enable output on PB1, PB2
// leave timer stopped (no input clock), set TCCR1B |= _BV(CS10) to start

static void InitTimer1(void)
{
    //do not use interrupts
    TIMSK1 = 0x00;
    //no input clock, timer stopped
    TCCR1B = 0x00;
    //reset timer
    TCNT1H = 0x00; //(buffered: write H first)
    TCNT1L = 0x00;

    //fast PWM, count=256 (mode 5), no prescaler,
    //enable compare outputs OC1A, OC1B in non-inverting mode
    TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM10);
    TCCR1B = _BV(WGM12);
    TCCR1C = 0x00;

    //set minimum pulse width for both outputs
    //note: counter is 16-bit, but only 8 bits are used; H byte is always 0x00
    OCR1AH = 0x00; //(buffered: write H first) this H byte will never be changed
    OCR1AL = 0x00;
    OCR1BH = 0x00; //this H byte will never be changed
    OCR1BL = 0x00;

    //enable PWM output on PB1, PB2
    DDRB |= (_BV(DDB1) | _BV(DDB2));

    //(timer is stopped and reset)
}

//-------------------------------------------------------------------------
// setup timer 2 for fast PWM, TOP = 0xff (cycle=256)
// set zero output, enable output on PD6, PD5
// leave timer stopped (no input clock), set TCCR2B |= _BV(CS20) to start

static void InitTimer2(void)
{
    //do not use interrupts
    TIMSK2 = 0x00;
    // use internal clock
    ASSR = 0x00;
    //no input clock, timer stopped
    TCCR2B = 0x00;
    //reset timer
    TCNT2 = 0x00;

    //fast PWM, count=256 (mode 3), no prescaler,
    //enable compare outputs OC2A, OC2B in non-inverting mode
    TCCR2A = _BV(COM2A1) | _BV(COM2B1) | _BV(WGM21) | _BV(WGM20);

    //set minimum pulse width for both outputs
    OCR2A = 0x00;
    OCR2B = 0x00;

    //enable PWM output on PB3, PD3
    DDRB |= _BV(DDB3);
    DDRD |= _BV(DDD3);

    //(timer is stopped and reset)
}

//-------------------------------------------------------------------------
// wait for the given number of PWM cycles

static void WaitPWMCycles(uint16_t cycles)
{
    while (cycles--)
    {
        SyncToPWM();
    }
}

//-------------------------------------------------------------------------
