//-------------------------------------------------------------------------
// play-v6 - music player for ATmega328P @ 16 MHz
// audio PWM control
// by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

#ifndef PWM_H
#define PWM_H

#include "userial.h"

//-------------------------------------------------------------------------
// test #defines

//set this to 1 to disable PWM sync and initial ramp-up for easier cycle counting
//(mainly useful for simulator)
#define TEST_MODE_NOSYNC 0

//-------------------------------------------------------------------------
// constants

// --- warning: changing CPU_CLOCK and/or PWM_CYCLE requires to regenerate
// --- tables in program memory and may also have unintended consequences

//CPU clock (Hz)
#define CPU_CLOCK               16000000ul
//PWM cycle length in clock cycles
#define PWM_CYCLE               256u

//analog zero output level for 8-bit phase-correct mode
#define ZERO_LEVEL              0x80

//inter-timer start delay, in CPU cycles
//(must be less than twice the cycle count min of UpdateVoiceSample() in play.c)
#define PWM_STAGGER_CYCLES      60u

//convert time in ms to PWM cycles (approx.), for use in constants
//(beware of overflow when storing the value in a 16-bit variable)
#define MS_TO_PWM_CYCLES(ms)    (((ms) * (CPU_CLOCK / 1000)) / PWM_CYCLE)

//ramp-up time to ZERO_LEVEL to avoid turn-on noise
#define RAMP_UP_TIME_MS         400u

//-------------------------------------------------------------------------
// PWM comparators for setting output values

#define PWM1Set     OCR0A       //PD6 (pin 6)
#define PWM2Set     OCR0B       //PD5 (pin 5)
#define PWM3Set     OCR1AL      //PB1 (pin 9)
#define PWM4Set     OCR1BL      //PB2 (pin 10)
#define PWM5Set     OCR2A       //PB3 (pin 11)
#define PWM6Set     OCR2B       //PD3 (pin 3)

//-------------------------------------------------------------------------
// macro: prepare to wait for sync by resetting PWM timer 0 overflow flag
//        (call immediately before SyncToPWM to be sure to sync exactly)

#define ResetPWMSync() \
do { \
    /*do not use '|' here*/ \
    TIFR0 = _BV(TOV0); \
} while(0)

//-------------------------------------------------------------------------
// macro: wait for end of PWM cycle, clear PWM timer 0 overflow flag
// used to have a full PWM cycle available to set a new value
// the test mode (cycle counting) version does not wait for end of cycle

#if TEST_MODE_NOSYNC

//non-stop version for tests
#define SyncToPWM() \
do { \
    /*always false, but useful for examining produced code*/ \
    while (TIFR0 == 0xff) \
    { } \
    /*reset count to 2 (found to be reasonably accurate for simulation)*/ \
    TCNT0 = 0x02; \
} while(0)

#else

//normal version for actual code
#define SyncToPWM() \
do { \
    while ((TIFR0 & _BV(TOV0)) == 0x00) \
    { } \
    /*clear PWM timer overflow flag for next call*/ \
    ResetPWMSync(); \
} while(0)

#endif //(! TEST_MODE_NOSYNC)

//-------------------------------------------------------------------------
// debug: keep track of the max. count reached by TCNT0 (for test measurements)
// if executing on target, send to UART0 when exceeded
// if emulating (TEST_MODE_NOSYNC), put a breakpoint where indicated
// (only compiled in debug mode)
// NOTE: do not use it more than once, or it will be compiled as a function

static __inline__ void DebugSendMaxTimer0(void)
{
    #ifdef DEBUG
    //volatile prevents this to be optimized out
    static volatile uint8_t maxCount = 0;

    //normal version for target tests
    uint8_t count = (uint8_t)TCNT0;
    if (count > maxCount)
    {
        //BRKP: in simulator, put a breakpoint here or watch maxCount
        maxCount = count;
        //if on target, send count as raw byte via UART0
        #if ! TEST_MODE_NOSYNC
        UART0_SendByte(maxCount);
        SyncToPWM();
        #endif
    }
    #endif
}

//-------------------------------------------------------------------------
// public functions

extern void InitAudioPWM(void);
extern void RampUpToSilence(void);

//-------------------------------------------------------------------------

#endif //PWM_H