//-------------------------------------------------------------------------
// minimalist, very fast, non-irq handler for UART0 of ATmega328P
// created 9 Dec 2013 by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

// NOTE: currently no error checking at all

#ifndef USERIAL_H
#define USERIAL_H

#include <stdbool.h>
#include <avr/io.h>

// +++ use setbaud.h?

//-------------------------------------------------------------------------
// macro to precompute baudFactor argument of UART0_Init

#define UART0_BAUD_FACTOR(cpuClock, baudRate)  ((cpuClock / 16 / baudRate) - 1)

//-------------------------------------------------------------------------
// public functions (also see fast macros below)

extern void    UART0_Init(uint16_t baudRateDivider);
extern void    UART0_SendByte(uint8_t byte);
extern uint8_t UART0_WaitByte(void);

//-------------------------------------------------------------------------
// summary (C equivalent) of the fast macros defined below
//
//   bool UART0_IsTxBufferFree(void);
//   bool UART0_IsRxDataAvailable(void);
//   void UART0_FlushRx(void);
//
//   void UART0_SendByteNoCheck(uint8_t byte); [*]
//   uint8_t UART0_ReceiveByteNoCheck(void);   [*]
//
// [*] direct access, potentially unsafe;
// when in doubt use regular functions from userial.c instead

//-------------------------------------------------------------------------
// true if tx buffer is empty

#define UART0_IsTxBufferFree()      ((UCSR0A & _BV(UDRE0)) != 0x00)

//-------------------------------------------------------------------------
// true if there is unread data in the receive buffer

#define UART0_IsRxDataAvailable()   ((UCSR0A & _BV(RXC0)) != 0x00)

//-------------------------------------------------------------------------
// clear receive (hardware) buffer

#define UART0_FlushRx() \
do { \
    while (UART0_IsRxDataAvailable()) \
    { \
        UART0_ReceiveByteNoCheck(); \
    } \
} while(0)

//-------------------------------------------------------------------------
// send data byte without checking first if tx buffer is free
// (non-blocking but unsafe, check with UART0_IsTxBufferFree() before calling)

#define UART0_SendByteNoCheck(byte) (UDR0 = (byte))

//-------------------------------------------------------------------------
// get byte from receive (hardware) buffer, return it
// (non-blocking but be sure to check with UART0_IsRxDataAvailable() before calling)

#define UART0_ReceiveByteNoCheck()  UDR0

//-------------------------------------------------------------------------

#endif //USERIAL_H
