---------------------------------------------------------------------------
-- music compiler for play-v6
-- by Enrico Colombini <erix@erix.it>
-- see musicomp.lua for license
--
-- module: code generator
-- see table at end of file for exported functions
---------------------------------------------------------------------------

-- access to public variables
local ipairs = ipairs
local assert = assert
local error = error
local floor = math.floor
local bor = bit32.bor
local lshift = bit32.lshift
local string = string
local io = io

-- all the rest is private
local privateEnv = {}
_ENV = privateEnv

---------------------------------------------------------------------------
-- constants

-- this could be raised to 8 in the current music data format
maxVoice = 6

---------------------------------------------------------------------------
-- code format

-- 'code' is a list of code bytes
-- code bytes are stored as Lua numbers, lower 8 bits are significant

---------------------------------------------------------------------------
-- add code bytes for NOTE command (start note) to code list
-- object format:
--   01nn nnnn   start playing note n (0..63) - similar to MIDI but 0 = C1
--   vvvv vsss   with volume v (0..31) on string/voice s (0..5)

function CodeStartNote(code, action)
    local noteIdx = action.note
    local voice = action.voice
    local volume = action.volume
    assert((noteIdx >= 0) and (noteIdx <= 63))
    assert((voice >= 1) and (voice <= maxVoice))
    assert((volume >= 0.0) and (volume <= 1.0))

    -- scale volume from 0.0..1.0 to 0..31 (approx. to nearest)
    volume = floor ((31.0 * volume) + 0.5)
    assert((volume >= 0) and (volume <= 31))
    
    code[#code + 1] = 0x40 + noteIdx
    code[#code + 1] = bor((voice - 1), lshift(volume, 3))
end

---------------------------------------------------------------------------
-- add code bytes for voice RELEASE command (release note) to code list
-- object format:
--   1000 0sss   stop playing note on string/voice s (release into silence)

function CodeReleaseNote(code, action)
    local voice = action.voice
    assert((voice >= 1) and (voice <= maxVoice))

    code[#code + 1] = 0x80 + (voice - 1)
end

---------------------------------------------------------------------------
-- add code bytes for WAIT command(s) to code list for the given duration,
-- do nothing is duration = 0
-- object format:
--   00tt tttt   wait for t 1/64 (0 = 64, else 1..63)

function CodeWait(code, action)
    local duration = action.duration
    assert (duration >= 0)

    -- max duration is 64/64, so split if necessary
    while duration >= 64 do
        code[#code + 1] = 0x00 -- (wait 64)
        duration = duration - 64
    end

    -- add residual (< 64) duration, if any
    if duration > 0 then
        code[#code + 1] = duration
    end
end

---------------------------------------------------------------------------
-- add code bytes for TEMPO command to code list (in 1/4 notes per minute)
-- object format:
--   1100 0001   set tempo to t (1/4 notes) per minute (1..255)
--   tttt tttt   values of t lesser than minimum are interpreted as minimum

function CodeTempo(code, action)
    local qpm = action.qpm
    
    assert((qpm >= 1) and (qpm <= 255))

    code[#code + 1] = 0xc1
    code[#code + 1] = qpm
end

---------------------------------------------------------------------------
-- add code bytes for INSTRUMENT command to code list
-- object format:
--   1100 0010   select instrument i (0..15) for voice s (0..5)
--   iiii 0sss

function CodeSelectInstrument(code, action)
    local instrument = action.instrument
    local voice = action.voice

    assert((instrument >= 0) and (instrument <= 15))
    assert((voice >= 1) and (voice <= maxVoice))

    code[#code + 1] = 0xc2
    code[#code + 1] = bor((voice - 1), lshift(instrument, 4))
end

---------------------------------------------------------------------------
-- add code byte for END command (end music) to code list
-- object format:
--   1100 0000   end music

function CodeEndMusic(code, action)
    code[#code + 1] = 0xc0
end

---------------------------------------------------------------------------
-- convert action stream to list of code bytes
-- return code list

function GenerateCode(actionStream)
    -- 'no-operation'
    local nop = {}

    -- action dispatch table
    -- functions are called as: f(code, action)
    local codeGenerators = {
        start_note = CodeStartNote,
        release_note = CodeReleaseNote,
        wait = CodeWait,
        tempo = CodeTempo,
        select_instrument = CodeSelectInstrument,
        end_music = CodeEndMusic,
        -- ignore directives already handled in stream processing
        -- (could be better to remove them in m_astream?)
        time_signature = nop,
        main_volume = nop,
        end_measure = nop,
        end_section = nop,    
        voice_mix = nop,
        triplet_1 = nop,
    }

    -- list of code bytes
    local code = {}
    
    -- scan action stream
    for _, action in ipairs(actionStream) do
        local act = action.action
            
        -- dispatch to appropriate code generator function
        local gen = codeGenerators[act]
        if not gen then
            error('codegen: unrecognized action: ' .. act, 0)
        elseif gen ~= nop then
            gen(code, action)
        end
    end
    
    return code
end

---------------------------------------------------------------------------
-- print code bytes (for testing)

function PrintCode(code, bytesPerLine)
    -- default
    bytesPerLine = bytesPerLine or 8
    
    local count = bytesPerLine
    for _, b in ipairs(code) do
        if count <= 0 then
            io.write('\n')
            count = bytesPerLine
        end
        
        io.write(string.format('0x%02x, ', b))
        count = count - 1
    end
    
    io.write('\n')
end

---------------------------------------------------------------------------
-- functions exported by this module 
---------------------------------------------------------------------------

return {
    GenerateCode = GenerateCode,
    PrintCode = PrintCode,
}

---------------------------------------------------------------------------
