---------------------------------------------------------------------------
-- music compiler for play-v6
-- by Enrico Colombini <erix@erix.it>
-- see musicomp.lua for license
--
-- module: compiler to event list
-- see table at end of file for exported functions
---------------------------------------------------------------------------

-- access to public variables
local ipairs = ipairs
local tonumber = tonumber
local pcall = pcall
local error = error
local io = io

-- all the rest is private
local privateEnv = {}
_ENV = privateEnv

---------------------------------------------------------------------------
-- event list format:

--   an event is a list of actions,
--   only a single event exists at a given 1/64 timestamp:
--     timestamp = { action, action, ... },
--
--   each action has a name (e.g. 'start_note') and zero or more named args:
--     { action=name, argname=argname, ... }
--   the last action is 'end_section' and its timestamp is the section duration

-- note: could document actions here, see m_codegen for list

---------------------------------------------------------------------------
-- constants (including those created during module initialization)

-- lowercase note names (English and Latin) from position in octave
noteNames = {
    [1]  = { 'c',  'do' },
    [2]  = { 'c#', 'do#',  'db', 'reb' },
    [3]  = { 'd',  're' },
    [4]  = { 'd#', 're#',  'eb', 'mib' },
    [5]  = { 'e',  'mi' },
    [6]  = { 'f',  'fa' },
    [7]  = { 'f#', 'fa#',  'gb', 'solb' },
    [8]  = { 'g',  'sol' },
    [9]  = { 'g#', 'sol#', 'ab', 'lab' },
    [10] = { 'a',  'la' },
    [11] = { 'a#', 'la#',  'bb', 'sib' },
    [12] = { 'b',  'si' },
}

-- inited with compiled note values from lowercase note name,
-- from 'C1'/'Do1'=0 to 'D#8'/'Re#8'=maxNoteValue+maxNoteTranspose
noteValues = {}
maxNoteValue = 63
maxNoteTranspose = 24

-- note durations in 1/64 from music notation value (e.g. 4 = 1/4 = 64/4)
-- including triplets approximated to 1/64
-- (could precompute but these are the only allowed values)
noteDurations = {
    -- base durations
    ['1'] = 64, ['2'] = 32, ['4'] = 16,
    ['8'] = 8, ['16'] = 4, ['32'] = 2, ['64'] = 1,
    -- triplets from 1/1 to 1/8 (too much timing error after that)
    ['1t1'] = 24, ['1t2'] = 20, ['1t3'] = 20,
    ['2t1'] = 12, ['2t2'] = 10, ['2t3'] = 10,
    ['4t1'] = 6, ['4t2'] = 5, ['4t3'] = 5,
    ['8t1'] = 3, ['8t2'] = 3, ['8t3'] = 2,
}

-- relative volume from 0.0 to 1.0
volumeByName = {
    fff = 1.0,
    ff = 0.875,
    f = 0.75,
    mf = 0.625,
    mp = 0.5,
    p = 0.375,
    pp = 0.25,
    ppp = 0.125,
}
-- initial volume if not given
defaultVolume = volumeByName['mf']

-- time signatures
simpleTimes = {
    '2/2', '2/4', '2/8', '3/2', '3/4', '3/8', '4/2', '4/4', '4/8', 'c', 'ct'
}
compoundTimes = {
    '6/4', '6/8', '6/16', '9/4', '9/8', '9/16', '12/4', '12/8', '12/16'
}

-- inited with all time signatures as keys, 'simple' or 'compound' as value
timeSignatures = {}

-- min, max tempo (in 1/4 notes per minute)
minTempoQpm = 10 -- note: somewhat arbitrary, should check on firmware
maxTempoQpm = 255

-- codes of instruments in the form { name=code, ... }
instrumentCodes = {}
-- name of file containing list of instruments
instrumentListFilename = 'instruments.list'
-- maximum number of instruments in list
maxInstrumentCount = 16

---------------------------------------------------------------------------
-- initialization, called at module loading (see end of file)
---------------------------------------------------------------------------

-- prepare lookup table from lowercase note name to compiled note value
-- from 'c1'/'do1'=0 to 'd#6'/'re#6'=63 (maxNoteValue)

function InitNoteValues()
    -- val=0 means C1/Do1
    local octave = 1
    local octaveIdx = 1

    for value = 0, maxNoteValue + maxNoteTranspose do
        -- for all names of this note
        for _, name in ipairs(noteNames[octaveIdx]) do
            -- build composite name, add it to table
            noteValues[name .. octave] = value
        end

        -- next note (semitone), wrap around at octave change
        octaveIdx = octaveIdx + 1
        if octaveIdx > 12 then
            octaveIdx = 1
            octave = octave + 1
        end
    end
end

---------------------------------------------------------------------------
-- prepare lookup table from time signature to either 'simple' or 'compound'

function InitTimeSignatures()
    for _, st in ipairs(simpleTimes) do
        timeSignatures[st] = 'simple'
    end

    for _, st in ipairs(compoundTimes) do
        timeSignatures[st] = 'compound'
    end
end

---------------------------------------------------------------------------
-- end of initialization
---------------------------------------------------------------------------

-- read instrument codes from file from given directory (called after init)
-- add name=code entries to instrumentCodes (codes start from 0)

function ReadInstrumentCodes(directory)
    local code = 0
    for name in io.lines(directory .. instrumentListFilename) do
        -- consider non-empty lines
        name = name:match('(%S*)%s*') -- remove terminator, if any (linux/win)
        if name then
            if code >= maxInstrumentCount then
                error('too many instruments in ' .. instrumentListFilename
                      .. ' (max=' .. maxInstrumentCount .. ')', 0)
            else
                instrumentCodes[name] = code
                code = code + 1
            end
        end
    end
end

---------------------------------------------------------------------------
-- add action to event at current timestamp,
-- create timestamp event if not already present,
-- an actions has the form { action=name, argname=argname, ... }
-- (a 'timestamp' field will be added by the stream processor)

function AddActionAtTimestamp(status, action)
    local eventList = status.events
    local timestamp = status.timestamp

    -- look for event at timestamp
    local event = eventList[timestamp]

    if event then
        -- an event already exists at this timestamp, add the action
        event[#event + 1] = action
    else
        -- no event at this timestamp yet, create one with the action
        eventList[timestamp] = { action }
    end
end

---------------------------------------------------------------------------
-- check for global source command: must be in voice 1

function CheckGlobalCmd(cmd, voiceNum)
    if voiceNum ~= 1 then
        error('\'' .. cmd .. '\' is only allowed in voice 1', 0)
    end
end

---------------------------------------------------------------------------
-- convert note/pause duration from source format to number of 1/64
-- source format can be:
--   music value, e.g. '4' = 1/4 = 16(/64)
--   dotted music value, e.g. '4..' = 1/4 + 1/8 + 1/16 = 28(/64)
--   triplet music value, e.g. '4t1', '4t2', '4t3': total = 1/4 = 16(/64)
-- return: 1) duration in 1/64
--         2) true if source ends with 'legato' symbol ('_')
--         3) 1, 2 or 3 as number if part of a triplet, else nil (no value)

function DurationTo64(src)
    -- check for trailing 'legato'
    local legato = false
    if src:sub(-1) == '_' then
        legato = true
        src = src:sub(1, #src - 1)
    end

    -- a duration is an alphanumeric string possibly followed by one or more dots
    -- (a string because it may be '4t1' etc. for triplets)
    local numstr, dotstr = src:match('^(%w+)(%.*)$')

    -- get base duration in 1/64 (the value must be in the duration table)
    local d = noteDurations[numstr]
    if not d then
        error('invalid duration: ' .. src, 0)
    end

    -- get triplet number, if any (else triplet = nil)
    local triplet = numstr:match('t([123])$')
    triplet = triplet and tonumber(triplet)

    -- start with base duration
    local duration = d

    -- for any dot, add halved duration (dotstr will be empty if no dots)
    for n = 1, #dotstr do
        d = d / 2
        if d < 1 then
            error('too many dots in duration: ' .. token, 0)
        else
            duration = duration + d
        end
    end -- (dots handling)

    return duration, legato, triplet
end

---------------------------------------------------------------------------
-- compile a pause into event list, if required
-- source format: p:duration
-- to: { action = 'release_note', voice = n, triplet = nil|1|2|3 }
-- if in position 1 of triplet, add: { action = 'triplet_1' }
-- update status.timestamp
-- note: see also CompileBlock

function CompilePause(token, cmd, arg, status)
    if status.legatoNote then
        error('legato cannot end in pause: ' .. token, 0)
    elseif arg == '' then
        error('missing duration: ' .. token, 0)
    end

    -- insert 'release note' command if enabled, unless it was already in pause
    if (not status.inPause) and status.releaseOnPause then
        AddActionAtTimestamp(status,
            { action = 'release_note', voice = status.voiceNum } )
    end

    -- convert arg to duration in 1/64, check for trailing legato and triplet
    local duration, trailingLegato, triplet = DurationTo64(arg)
    if trailingLegato then
        error('pause cannot start a legato: ' .. token, 0)
    end

    -- if triplet in position 1, add reference (for accent computations)
    if triplet == 1 then
        AddActionAtTimestamp(status, { action = 'triplet_1' } )
    end

    -- advance timestamp to end of pause
    status.timestamp = status.timestamp + duration

    -- remember pause status, to avoid unneeded release commands
    status.inPause = true
end

---------------------------------------------------------------------------
-- compile a note into event list (unless leading legato from same note)
-- source format: [_]notename:duration[_]
-- to: { action = 'start_note', note = noteval, voice = n, triplet = nil|1|2|3 }
-- if in position 1 of triplet, add in any case: { action = 'triplet_1' }
-- update status.timestamp

function CompileNote(token, cmd, arg, status)
    -- check for legato continuation, get note value
    local leadingLegato, note = cmd:match('^(%_?)(%S+)$')
    leadingLegato = (leadingLegato ~= '')

    local noteval = noteValues[note]
    if not noteval then
        error('invalid note: ' .. token, 0)
    elseif arg == '' then
        error('missing duration: ' .. token, 0)
    end

    -- check and apply transpose
    noteval = noteval + status.transpose
    if (noteval < 0) or (noteval > maxNoteValue) then
        error('transposed note out of range: ' .. token, 0)
    end

    -- legato continuation means there must be a pending legato
    if leadingLegato and (not status.legatoNote) then
        error('no previous legato: ' .. token, 0)
    elseif status.legatoNote and (not leadingLegato) then
        error('note should continue legato: ' .. token, 0)
    end

    -- convert arg to duration in 1/64, check for trailing legato and triplet
    local duration, trailingLegato, triplet = DurationTo64(arg)

    if leadingLegato then
        if note == status.legatoNote then
            -- legato on same note: do nothing (will just add to duration)
        else
            -- legato on different note (implement in firmware first)
            error('legato on different note not implemented yet', 0)
        end
    else
        -- no leading legato, start note with current timestamp, add object code to events
        AddActionAtTimestamp(status,
            { action = 'start_note', note = noteval,
              voice = status.voiceNum, triplet = triplet } )
    end

    -- remember this note if it started (or continues) a legato
    if trailingLegato then
        status.legatoNote = note
    else
        status.legatoNote = nil
    end

    -- triplet in position 1, even if no note action generated:
    -- add reference for accent computations
    if triplet == 1 then
        AddActionAtTimestamp(status, { action = 'triplet_1' } )
    end

    -- advance timestamp to end of note
    status.timestamp = status.timestamp + duration

    -- the first pause will (possibly) terminate this note
    status.inPause = false
end

---------------------------------------------------------------------------
-- compile a global tempo setting into event list
-- source format: tempo:ref=rpm (reference durations per minute),
-- usually ref corresponds to the beat time, e.g. tempo:4.=60 (compound time)
-- to: { action = 'tempo', qpm = quartersPerMinute }

function CompileTempo(token, cmd, arg, status)
    CheckGlobalCmd(cmd, status.voiceNum) -- (error if check fails)

    -- get reference duration number of references per minute
    local ref, rpm = arg:match('^(%d+%.?)%=(%d+)$')
    if not (ref and rpm) then
        error('invalid tempo notation: ' .. token, 0)
    end

    -- convert reference duration to 1/64 (ignore extra results)
    ref64 = DurationTo64(ref)
    rpm = tonumber(rpm)
    if (rpm <= 0) then
        error('tempo value must be a positive integer: ' .. token, 0)
    end

    -- convert to qpm (number of 1/4 notes, or 16/64, per minute)
    local qpm = rpm * ref64 / 16

    -- clamp to valid range
    if qpm < minTempoQpm then
        io.stderr:write('- warning: tempo raised from ', qpm,
                        ' to ', minTempoQpm, ' qpm: ', token, '\n')
        qpm = minTempoQpm
    elseif qpm > maxTempoQpm then
        io.stderr:write('- warning: tempo reduced from ', qpm,
                        ' to ', maxTempoQpm, ' qpm: ', token, '\n')
        qpm = maxTempoQpm
    end

    -- add action
    AddActionAtTimestamp(status, { action = 'tempo', qpm = qpm } )
end

---------------------------------------------------------------------------
-- compile time signature into event list
-- source format: time:n/u (number of units / unit), e.g. 3/8
-- special notations are converted: 'c' -> '4/4', 'ct' -> 2/2
-- to: { action = 'time_signature', time = 'n/u', compound = bool }

function CompileTime(token, cmd, arg, status)
    CheckGlobalCmd(cmd, status.voiceNum) -- (error if check fails)

    local timeType = timeSignatures[arg]
    if not timeType then
        error('unrecognized time signature: ' .. token, 0)
    end

    -- convert special notations to the n/u equivalent format
    if arg == 'c' then
        arg = '4/4'
    elseif arg == 'ct' then
        arg = '2/2'
    end

    -- add action with time as string, guaranteed to be in the form 'n/u'
    AddActionAtTimestamp(status,
        { action = 'time_signature', time = arg,
          compound = (timeType == 'compound') } )
end


---------------------------------------------------------------------------
-- compile main volume setting into event list
-- source format: vol:level (must be in volumeByName table, from ppp to fff)
-- to: { action = 'main_volume', volume = fpVolume }

function CompileVolume(token, cmd, arg, status)
    CheckGlobalCmd(cmd, status.voiceNum) -- (error if check fails)

    local volume = volumeByName[arg]
    if not volume then
        error('invalid volume name: ' .. token, 0)
    end

    -- main_volume range is 0.0 to 1.0
    AddActionAtTimestamp(status, { action = 'main_volume', volume = volume } )
end

---------------------------------------------------------------------------
-- compile voice mixer setting into event list
-- source format: mix:level% (from 0% to 100%)
-- to: { action = 'voice_mix', level = fpLevel, voice = n }

function CompileMixer(token, cmd, arg, status)
    local level = arg:match('^(%d+)%%$')

    if (not level) or (tonumber(level) > 100) then
        error('voice mix value must be from 0% to 100%: ' .. token, 0)
    end

    -- add action, 100% -> 1.0
    AddActionAtTimestamp(status,
        { action = 'voice_mix', level = tonumber(level) / 100.0,
          voice = status.voiceNum } )
end

---------------------------------------------------------------------------
-- compile instrument selection into event list
-- source format: instrum:name
-- to: { action = 'select_instrument', instrument = code , voice = n }

function CompileInstrument(token, cmd, arg, status)
    local iname = arg

    -- get numeric code
    local icode = iname and instrumentCodes[iname]

    if not icode then
        error('unknown instrument: ' .. iname, 0)
    end

    -- add action
    AddActionAtTimestamp(status,
        { action = 'select_instrument', instrument = icode,
          voice = status.voiceNum } )
end

---------------------------------------------------------------------------
-- compile transpose instruction (+/-n semitones):
-- set transpose value for voice (relative to current), no event is generated
-- source format: transpose:n (n can be positive, 0 or negative, no check here)

function CompileTranspose(token, cmd, arg, status)
    local n = arg:match('^([+-]?%d+)$')

    if not n then
        error('transpose must indicate a number of semitones: ' .. token, 0)
    end

    -- adjust current transpose value, will be checked later
    status.transpose = status.transpose + tonumber(n);
end

---------------------------------------------------------------------------
-- compile release instruction (on/off, default=on):
-- set releaseOnPause flag accordingly

function CompileRelease(token, cmd, arg, status)
    local t = { on = true, off = false }

    -- (note: cannot use boolean check here)
    if t[arg] == nil then
        error('release must be on/off: ' .. token, 0)
    end

    -- adjust current release status
    status.releaseOnPause = t[arg]
end

---------------------------------------------------------------------------
-- compile a single voice line
-- (a line could contain at most a single measure),
-- add actions to events list, indexed by timestamp (in 1/64)
-- update status: add any events, update to timestamp after line,
-- return duration (could be zero if there are no notes or pauses)

function CompileVoiceLine(source, status)
    -- command dispatch table
    -- functions are called as: f(token, cmd, arg, status)
    local commandCompilers = {
        p = CompilePause,       -- pause: end of note
        tempo = CompileTempo,   -- set tempo in qpm
        time = CompileTime,     -- set time signature
        volume = CompileVolume, -- set global volume (p, mf, f, ...)
        mix = CompileMixer,     -- set relative volume of voice (0%..100%)
        instrument = CompileInstrument, -- select instrument (0..maxInstrument)
        transpose = CompileTranspose, -- transpose following notes by +/-n semitones
        release = CompileRelease, -- set releaseOnPause on/off (default:on)
    }

    -- get initial timestamp
    local startTimestamp = status.timestamp

    -- scan source string
    for token in source:gmatch('%S+') do
        -- do it all in lowercase
        token = token:lower()
        -- examine token in the form cmd:arg (:arg is optional)
        local cmd, arg = token:match('^([^%:]+)%:?([^%:]*)$')

        if not cmd then
            error('invalid music notation: ' .. token, 0)

        elseif commandCompilers[cmd] then
            -- dispatch to appropriate function
            commandCompilers[cmd](token, cmd, arg, status)

        else
            -- must be a note (will be checked)
            CompileNote(token, cmd, arg, status)

        end -- (source token cases)

    end -- (source string scan)

    -- return duration
    return status.timestamp - startTimestamp
end

---------------------------------------------------------------------------
-- compile block of voice lines 
-- (representing a full or partial measure, or just some commands)
-- add actions to events list, update timestamp in all voices status
-- see m_loader for block format
-- return block duration (in 1/64), could be 0

function CompileBlock(section, block, voicesStatus)
    -- duration of this block (not determined yet, 0 is a valid duration)
    local blockDuration = nil
    local maxVoices = #voicesStatus
    
    -- compile voice lines for this block
    for voiceNum = 1, maxVoices do
        local voiceLine = block[voiceNum] -- may be nil
        local status = voicesStatus[voiceNum]

        local ok, r = pcall(
            function()
                -- check for empty line
                if voiceLine and (voiceLine ~= '-') then
                    -- normal line: compile it, add events to eventList 
                    -- (will update status, including timestamp)
                    local duration = CompileVoiceLine(voiceLine, status)

                    -- all defined line durations must have the same value
                    if blockDuration and (duration ~= blockDuration) then
                        error('duration is ' .. duration .. '/64 instead of '
                              .. blockDuration .. '/64', 0)
                    end

                    if not blockDuration then
                        -- first duration encountered
                        blockDuration = duration
                    end
                end
            end -- (function)
        ) -- (pcall)

        -- check pcall result
        if not ok then
            -- prepend context info to error message
            error('[' .. section.name .. '] line '
                  .. (block.linenum + voiceNum - 1) .. ': '.. r, 0)
        end
    end -- (voice lines)

    -- if no voices defined, no duration
    blockDuration = blockDuration or 0

    if blockDuration and (blockDuration > 0) then
        -- auto-fill empty lines with pause-equivalent
        for voiceNum = 1, maxVoices do
            local voiceLine = block[voiceNum] -- may be nil
            local status = voicesStatus[voiceNum]

            if (not voiceLine) or (voiceLine == '-') then
                -- insert 'release note' command if enabled, 
                -- unless it was already in pause or legato
                if (not status.inPause) and status.releaseOnPause 
                   and (not status.legato) then
                    AddActionAtTimestamp(status,
                        { action = 'release_note', voice = voiceNum } )
                    -- remember pause status, to avoid unneeded release commands
                    status.inPause = true
                end
                
                -- advance timestamp to be the same as non-empty voices
                status.timestamp = status.timestamp + blockDuration
            end
        end
    end
    
    -- if all voice lines are empty, return 0
    return blockDuration or 0
end

---------------------------------------------------------------------------
-- compile music section from section table into event list
-- a section is made of line blocks that may or may not represent measures
-- add: { action = 'end_measure', duration = duration } when needed
-- add final { action = 'end_section' }
-- return event list for section

function CompileSection(section, maxVoices)
    -- eventList: list of music events indexed by timestamp (in 1/64),
    -- each event is an ordered list of actions at that timestamp
    local eventList = {
        -- timestamp1 = { action1_voice1, action2_voice1, action1_voice2, ... },
        -- timestamp2 = { action1_voice1, action2_voice1, action1_voice2, ... },
    }

    -- prepare default status for all voices
    local voicesStatus = {}
    for voiceNum = 1, maxVoices do
        voicesStatus[voiceNum] = {
            events = eventList,     -- common event list (each voice adds to it)
            voiceNum = voiceNum,    -- for convenience
            releaseOnPause = true,  -- true = insert release commands, false = natural decay
            timestamp = 0,          -- from start of section, in 1/64
            inPause = true,         -- starts silent (used to join consecutive pauses)
            legatoNote = nil,       -- name of last note, if it had a trailing legato
            transpose = 0,          -- number of semitones to transpose notes
        }
    end

    for blockNum, block in ipairs(section) do
        -- compile block, add events to eventList
        local blockDuration = CompileBlock(section, block, voicesStatus)
    
        -- safety check: all timestamps must be in sync
        local timestamp1 = voicesStatus[1].timestamp
        for voiceNum = 2, maxVoices do
            if voicesStatus[voiceNum].timestamp ~= timestamp1 then
                error('[' .. section.name .. '] line '
                      .. (block.linenum + voiceNum - 1) 
                      .. ': compiler error: timestamp for voice ' 
                      .. voiceNum .. ' out of sync', 0)
            end
        end
        
        -- if block was a (full or partial) measure,
        -- add 'end of measure' action to last voice
        if blockDuration > 0 then
            AddActionAtTimestamp(voicesStatus[maxVoices],
                { action = 'end_measure', duration = blockDuration } )
        end

    end -- (blocks)


    -- add final timestamp to last voice to give section the correct length
    AddActionAtTimestamp(voicesStatus[maxVoices], { action = 'end_section' } )

    -- list of events indexed by timestamp (unordered)
    return eventList
end

---------------------------------------------------------------------------
-- init when loading module
---------------------------------------------------------------------------

InitNoteValues()
InitTimeSignatures()

---------------------------------------------------------------------------
-- functions exported by this module
---------------------------------------------------------------------------

return {
    defaultVolume = defaultVolume,
    ReadInstrumentCodes = ReadInstrumentCodes,
    CompileSection = CompileSection,
}

---------------------------------------------------------------------------
