---------------------------------------------------------------------------
-- music compiler for play-v6
-- created Jan 2014 by Enrico Colombini <erix@erix.it>
-- see musicomp.lua for license
--
-- module: output file writer
-- see table at end of file for exported functions
---------------------------------------------------------------------------

-- access to public variables
local ipairs = ipairs
local bxor = bit32.bxor
local band = bit32.band
local rshift = bit32.rshift
local char = string.char
local io = io

-- all the rest is private
local privateEnv = {}
_ENV = privateEnv

---------------------------------------------------------------------------
-- output binary file format:

-- 'L', dataLenL, dataLenH, code, CRC-8
-- (CRC=8 includes dataLenL, dataLenH, code)

---------------------------------------------------------------------------
-- compute 'iButton' (Dallas/Maxim) CRC-8 from previous CRC-8
-- and data value representing a byte (adapted from avr-libc manual)
-- return new CRC-8 value

function Crc8Update(crc, data)
    crc = bxor(crc, data)
    for i = 0, 7 do
        if (band(crc, 0x01) ~= 0) then
            crc = bxor(rshift(crc, 1), 0x8c)
        else
            crc = rshift(crc, 1);
        end
    end

    return crc
end

---------------------------------------------------------------------------
-- write binary output file from code (list of numbers representing bytes)
-- no check for existing file before overwriting

function WriteOutputFile(filename, code)
    -- create output file
    local outf = io.open(filename, 'wb')

    if outf then
        -- LOAD command
        outf:write('L')

        -- CRC-8 includes all bytes from now on
        local crc8 = 0

        -- data length (& add to CRC)
        local lenL = band(#code, 0xff)
        local lenH = rshift(#code, 8)
        crc8 = Crc8Update(crc8, lenL)
        crc8 = Crc8Update(crc8, lenH)
        outf:write(char(lenL))
        outf:write(char(lenH))

        -- code bytes (& add to CRC)
        for _, b in ipairs(code) do
            crc8 = Crc8Update(crc8, b)
            outf:write(char(b))
        end

        -- CRC-8
        outf:write(char(crc8))

        -- done
        outf:close();
    end
end

---------------------------------------------------------------------------
-- functions exported by this module 
---------------------------------------------------------------------------

return {
    WriteOutputFile = WriteOutputFile,
}

---------------------------------------------------------------------------

